<?php
namespace AutomateYourTests\Model\Config;

/**
 * This class validates that the test suite syntax is valid. EVERY unexpected instruction that
 * the software can not handle is protected against here upfront BEFORE any tests are executed.
 *
 * @link        http://automateyourtests.com
 * @version     1.2.2
 * @author      Marc Daniel Dammann <howto@automateyourtests.com>
 * @copyright   2016 Marc Daniel Dammann
 * @license     GPL-3.0 <https://opensource.org/licenses/GPL-3.0>
 *
 * Copyright 2016 Marc Daniel Dammann
 * This file is part of AutomateYourTests. AutomateYourTests is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version. AutomateYourTests is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with AutomateYourTests. If not, see http://www.gnu.org/licenses/.
 */

class Validator extends Base {
    /**
     * Class vars
     */
    protected $validationErrors = array();

    /**
     *
     */
    public function checkSome($config = null)
    {
        // TODO: WORKAROUND: Pro-forma root to recurse over
        $configRoot = new \stdClass();
        $configRoot->testsuite = $config;

        $this->checkNodeStructure($configRoot, $this->validXmlStructure);
        $this->checkDataType($configRoot, $this->validXmlStructure);
        $this->checkOccurances($configRoot, $this->validXmlStructure);
        $this->checkRequired($configRoot, $this->validXmlStructure);
        $this->checkEmpty($configRoot);

        $this->reportValidationErrors();
    }

    /**
     *
     */
    public function checkAll($config = null)
    {
        // TODO: WORKAROUND: Pro-forma root to recurse over
        $configRoot = new \stdClass();
        $configRoot->testsuite = $config;

        //$this->checkNodeStructure($configRoot, $this->validXmlStructure);

        // Do this validation once again, who knows what happened in the meantime?
        // TODO: But aren't the reported line numbers wrong now, cos the XML has been transformed?
        $this->checkSome($config);

        // Check tag dependencies that otherwise would NOT be 'required'
        $this->checkRequiredCustomRules($configRoot, 'user', '', 'url');
        $this->checkRequiredCustomRules($configRoot, 'post', '', 'url');
        $this->checkRequiredCustomRules($configRoot, 'contains', '', 'url');
        $this->checkRequiredCustomRules($configRoot, 'notcontains', '', 'url');
        $this->checkRequiredCustomRules($configRoot, 'xpath', '', 'url');
        $this->checkRequiredCustomRules($configRoot, 'status', '', 'url');
        $this->checkRequiredCustomRules($configRoot, 'notstatus', '', 'url');

        // Check URL is valid syntax
        $this->checkUrlSyntax($configRoot, $this->validXmlStructure);

        // Validate logins settings
        $this->checkLoginCookies($configRoot);

        // Validate attributes
        $this->checkUnknownAttributes($configRoot, $this->validXmlStructure);
        $this->checkRequiredAttributes($configRoot, $this->validXmlStructure);

        // Validate labels
        $this->checkUserLabels($configRoot);
        $this->checkSetupLabels($configRoot);
        $this->checkTeardownLabels($configRoot);
        $this->checkDuplicateLabels($configRoot);

        $this->reportValidationErrors();
    }

    /**
     *
     */
    protected function checkNodeStructure($realNodes, $validNodes)
    {
        foreach ($realNodes as $realNode) {
            if ($this->isRealNodeInPlace($realNode, $validNodes)) {
                $validNodeChildren = $this->getNextValidNodes($realNode, $validNodes);

                if (is_array($validNodeChildren)) {
                    $this->checkNodeStructure($realNode, $validNodeChildren);
                } else {
                    $validBottomKeys = array_keys($validNodes);
                    $realBottomKeys = array_keys((array)$realNodes);

                    foreach ($validBottomKeys as $validBottomKey) {
                        if (! in_array($validBottomKey, $realBottomKeys)) {
                            $isBottomNodeRequired = $validNodes[$validBottomKey]['required'];

                            if ($isBottomNodeRequired) {
                                $this->addError($realNodes, 'The <' . $validBottomKey . '> node is required, but it currently does not exist or is empty');
                            }
                        }
                    }
                }
            } elseif (! $this->suggestNodes($realNode, $validNodes)) {
                $this->addError($realNode, 'The <' . $realNode->getName() . '> node is unknown, so please remove it');
            }
        }
    }

    /**
     *
     */
    protected function checkDataType($realNodes, $validNodes)
    {
        foreach ($realNodes as $realNode) {
            if ($this->isRealNodeInPlace($realNode, $validNodes)) {
                $validNodeChildren = $this->getNextValidNodes($realNode, $validNodes);

                if (is_array((array) $realNode) && count($realNode)) {
                    $realNodeDataType = 'array';
                } elseif (is_string((string) $realNode) && ! is_numeric($realNode)) {
                    $realNodeDataType = 'string';
                } elseif (is_int($realNode)) {
                    $realNodeDataType = 'integer';
                } else {
                    $realNodeDataType = 'unknown';
                }

                // Another validator tsake care of empty nodes
                if ($realNodeDataType != 'string' && $realNode != '') {
                    if ($realNodeDataType != gettype($validNodeChildren)) {
                        $this->addError($realNode, 'The <' . $realNode->getName() . '> is of type "' . $realNodeDataType . '", but must be of type "' . $validNodeChildren . '"  instead');
                    }
                }

                $this->checkDataType($realNode, $validNodeChildren);
            }
        }
    }

    /**
     *
     */
    protected function checkOccurances($realNodes, $validNodes)
    {
        $doubleNodeNameChecker = array();

        foreach ($realNodes as $realNode) {
            if ($this->isRealNodeInPlace($realNode, $validNodes)) {
                $validNodeChildren = $this->getNextValidNodes($realNode, $validNodes);
                $realNodeName = $realNode->getName();

                if (isset($validNodes[$realNodeName]) && $validNodes[$realNodeName]['occurance'] == 'scalar') {
                    if (\in_array($realNodeName, $doubleNodeNameChecker)) {
                        $this->addError($realNode, 'The <' . $realNodeName . '> node must only occur once and not multiple times');
                    }
                }

                $doubleNodeNameChecker[] = $realNodeName;

                $this->checkOccurances($realNode, $validNodeChildren);
            }
        }
    }

    /**
     *
     */
    protected function checkRequired($realNodes, $validNodes)
    {
        $realNodeNames = array();

        // What real nodes exist?
        foreach ($realNodes as $realNode) {
            $realNodeNames[] = $realNode->getName();
        }

        // Validate missing real nodes
        foreach ((array) $validNodes as $validNodeName => $validNode) {
            if (isset($validNode['required'])) {
                if ($validNode['required']) {
                    if (!in_array($validNodeName, $realNodeNames)) {
                        $this->addError($realNodes, 'The <' . $validNodeName . '> child node is missing');
                    }
                }
            }
        }

        // Validate missing real node values
        foreach ($realNodes as $realNode) {
            if ($this->isRealNodeInPlace($realNode, $validNodes)) {
                $validNodeChildren = $this->getNextValidNodes($realNode, $validNodes);
                $realNodeName = $realNode->getName();

                if (isset($validNodes[$realNodeName]['required'])) {
                    if (isset($validNodes[$realNodeName]) && $validNodes[$realNodeName]['required']) {
                        if (\is_array($validNodeChildren)) {
                            if (\count($realNode) == 0) {
                                $this->addError($realNode, 'The <' . $realNodeName . '> node must have at least one required child node');
                            }
                        } elseif (\trim($realNode) == '') {
                            $this->addError($realNode, 'The <' . $realNodeName . '> node must have a required value');
                        }
                    }
                }

                $this->checkRequired($realNode, $validNodeChildren);
            }
        }
    }

    /**
     *
     */
    protected function checkRequiredCustomRules($realNodes, $dependentTagKey, $dependentLabelKey, $dependerTagKey)
    {
        $realNodesByKey = array();

        // Real nodes by key
        foreach ($realNodes as $realNode) {
            $realNodesByKey[$realNode->getName()][$realNode->attribute('label')] = $realNode;
        }

        // A <post> needs a <url>
        if (isset($realNodesByKey[$dependentTagKey][$dependentLabelKey])) {
            if (! isset($realNodesByKey[$dependerTagKey])) {
                $this->addError($realNodesByKey[$dependentTagKey][$dependentLabelKey], 'The <' . $dependentTagKey . '> node has been specified, but there is no <' . $dependerTagKey . '> node');
            } elseif (\trim($realNodesByKey[$dependerTagKey][$dependentLabelKey]) == '') {
                $this->addError($realNodesByKey[$dependentTagKey][$dependentLabelKey], 'The <' . $dependentTagKey . '> node has been specified, but the <' . $dependerTagKey . '> node node is empty');
            }
        }

        foreach ($realNodes as $realNode) {
            $this->checkRequiredCustomRules($realNode, $dependentTagKey, $dependentLabelKey, $dependerTagKey);
        }
    }

    /**
     *
     */
    protected function checkEmpty($nodes)
    {
        foreach ($nodes as $node) {
            $nodeName = $node->getName();

            // TODO: Introduce an extra $this->validXmlStructure attribute for these exceptional tags (or introduce an extra test assertion attribute
            if ($nodeName != 'contains' && $nodeName != 'notcontains' && $nodeName != 'status' && $nodeName != 'notstatus' && $nodeName != 'xpath') {
                // WORKAROUND: I can't just test for whitespace in strings only, so this is to avoid SimpleXmlElement objects
                if (count($node) == 0 && strlen(trim($node)) == 0) {
                    $this->addError($node, 'The <' . $nodeName . '> node is empty. Did you forget to add content to it!?');
                }
            }

            $this->checkEmpty($node);
        }
    }

    /**
     *
     */
    protected function getValidAttributes($realNode, $validNodes)
    {
        $realNodeName = $realNode->getName();

        if (isset($validNodes[$realNodeName]['attributes'])) {
            $validAttributes = $validNodes[$realNodeName]['attributes'];
        // TODO: Multiple <userstory>, <given>, <when>, ... tags
        //} elseif ($realNodeName == 'userstory') {
        //    $validAttributes = $validNodes['attributes'];
        } else {
            $validAttributes = array();
        }

        return \array_keys($validAttributes);
    }

    /**
     *
     */
    protected function getValidRequiredAttributes($validNodes)
    {
        $requiredAttributes = array();

        //$realNodeName = $realNode->getName();

        if (isset($validNodes['attributes'])) {
            $validAttributes = $validNodes['attributes'];
        // TODO: Multiple <userstory>, <given>, <when>, ... tags
        //} elseif ($realNodeName == 'userstory') {
        //    $validAttributes = $validNodes['attributes'];
        } else {
            $validAttributes = array();
        }

        if ($validAttributes) {
            foreach ((array) $validAttributes as $attributeKey => $attribute) {
                if ($attribute['required']) {
                    $requiredAttributes[] = $attributeKey;
                }
            }
        }

        return $requiredAttributes;
    }

    /**
     *
     */
    protected function checkUrlSyntax($realNodes, $validNodes)
    {
        foreach ($realNodes as $realNode) {
            if ($this->isRealNodeInPlace($realNode, $validNodes)) {
                $validNodeChildren = $this->getNextValidNodes($realNode, $validNodes);

                if (! is_array($validNodeChildren)) {
                    $realNodeName = $realNode->getName();

                    if ($realNodeName == 'url') {
                        if (filter_var($realNode, FILTER_VALIDATE_URL) === false) {
                            $this->addError($realNode, 'The <' . $realNodeName . '> is set to "' . $realNode . '", which is not a valid URL format');
                        }
                    }
                }

                $this->checkUrlSyntax($realNode, $validNodeChildren);
            }
        }
    }

    /**
     *
     */
    protected function checkUnknownAttributes($realNodes, $validNodes)
    {
        foreach ($realNodes as $realNode) {
            if ($this->isRealNodeInPlace($realNode, $validNodes)) {
                $realAttributes = $realNode->attributes();
                $validAttributes = $this->getValidAttributes($realNode, $validNodes);

                foreach ($realAttributes as $key => $value) {
                    if (! \in_array($key, $validAttributes)) {
                        $this->addError($realNode, 'The <' . $realNode->getName() . '> node has a "' . $key . '" attribute, which is unknown');
                    }
                }

                $validNodeChildren = $this->getNextValidNodes($realNode, $validNodes);

                $this->checkUnknownAttributes($realNode, $validNodeChildren);
            }
        }
    }

    /**
     *
     */
    protected function checkRequiredAttributes($realNodes, $validNodes)
    {
        foreach ($realNodes as $realNode) {
            if ($this->isRealNodeInPlace($realNode, $validNodes)) {
                $realNodeName = $realNode->getName();

                $validRequiredAttributes = $this->getValidRequiredAttributes($validNodes[$realNodeName]);

                foreach ($validRequiredAttributes as $validRequiredAttribute) {
                    if (! $realNode->hasAttribute($validRequiredAttribute)) {
                        $this->addError($realNode, 'The <' . $realNode->getName() . '> node must have a required "' . $validRequiredAttribute . '" attribute');
                    }
                }

                $validNodeChildren = $this->getNextValidNodes($realNode, $validNodes);

                $this->checkRequiredAttributes($realNode, $validNodeChildren);
            }
        }
    }

    /**
     * TODO: Not ideal. We need this cookiedir="" check currently, because <logins> is not required
     */
    protected function checkLoginCookies($realNodes)
    {
        if ($realNodes->testsuite->logins) {
            if (empty($realNodes->testsuite->logins->attribute('cookiedir'))) {
                $this->addError($realNodes->testsuite->logins, 'The cookiedir="" attribute does not exist in <logins>');
            }

            //if (strpos($realNodes->testsuite->logins->cookies->file, '{USER_LABEL}') === false) {
            //    $this->addError($realNodes->testsuite->logins, 'The cookies file name must contain a {USER_LABEL} placeholder');
            //}
        }
    }

    /**
     *
     */
    protected function checkUserLabels($realNodes)
    {
        if ($realNodes->testsuite->logins) {
            $userLabels = $this->getLabels($realNodes, 'user', 'logins');

            $this->matchLabels($realNodes, $userLabels, 'user', 'logins');
        }
    }

    /**
     *
     */
    protected function checkSetupLabels($realNodes)
    {
        if ($realNodes->testsuite->fixture->setup) {
            $setupLabels = $this->getLabels($realNodes, 'setup', 'fixture');

            $this->matchLabels($realNodes, $setupLabels, 'setup', 'fixture');
        }
    }

    /**
     *
     *
     */
    protected function checkTeardownLabels($realNodes)
    {
        if ($realNodes->testsuite->fixture->teardown) {
            $teardownLabels = $this->getLabels($realNodes, 'teardown', 'fixture');

            $this->matchLabels($realNodes, $teardownLabels, 'teardown', 'fixture');
        }
    }

    /**
     *
     */
    protected function getLabels($realNodes, $labelNodeName, $parentNodeName)
    {
        $labels = array();

        // Get labels
        if ($realNodes->testsuite->$parentNodeName->$labelNodeName) {
            foreach ($realNodes->testsuite->$parentNodeName->$labelNodeName as $node) {
                $labels[] = $node->attribute('label');
            }
        }

        return $labels;
    }

    /**
     * TODO: Try to slim this method down like in the commit "Fix all remaining FIXME tests"
     */
    protected function checkDuplicateLabels($realNodes)
    {
        $duplicateLabels = array();

        // Get labels
        $userLabels = $this->getLabels($realNodes, 'user', 'logins');
        $setupLabels = $this->getLabels($realNodes, 'setup', 'fixture');
        $teardownLabels = $this->getLabels($realNodes, 'teardown', 'fixture');

        // Marge labels
        $labels = array_merge($userLabels, $setupLabels, $teardownLabels);

        // Prevent duplicates
        if ($realNodes->testsuite->logins) {
            foreach ($realNodes->testsuite->logins->user as $node) {
                if (in_array($node->attribute('label'), $duplicateLabels)) {
                    $this->addError($node, 'The <' . $node->getName() . ' label="' . $node->attribute('label') . '"> label is duplicated');
                }

                $duplicateLabels[] = $node->attribute('label');
            }
        }
        if ($realNodes->testsuite->fixture->setup) {
            foreach ($realNodes->testsuite->fixture->setup as $node) {
                if (in_array($node->attribute('label'), $duplicateLabels)) {
                    $this->addError($node, 'The <' . $node->getName() . ' label="' . $node->attribute('label') . '"> label is duplicated');
                }

                $duplicateLabels[] = $node->attribute('label');
            }
        }
        if ($realNodes->testsuite->fixture->teardown) {
            foreach ($realNodes->testsuite->fixture->teardown as $node) {
                if (in_array($node->attribute('label'), $duplicateLabels)) {
                    $this->addError($node, 'The <' . $node->getName() . ' label="' . $node->attribute('label') . '"> label is duplicated');
                }

                $duplicateLabels[] = $node->attribute('label');
            }
        }

        return $labels;
    }

    /**
     *
     */
    protected function matchLabels($realNodes, $labels, $labelNodeName, $parentNodeName)
    {
        foreach ($realNodes as $realNode) {
            if ($realNode->$labelNodeName && ! isset($realNodes->$parentNodeName)) {
                // Be lenient
                // TODO: Trim all labels across the board upfront only once - not several times!
                $trimmedLabelNodeName = trim($realNode->$labelNodeName);

                // WORKAROUND: Only end up here for <logins> and <fixture>
                if (! \in_array($trimmedLabelNodeName, $labels)) {
                    $this->addError($realNode->$labelNodeName, 'Theee <' . $labelNodeName . '> label "' . $realNode->$labelNodeName . '" does not exist in <' . $parentNodeName . '>');
                }
            } else {
                // WORKAROUND: Only end up here for <step>
                $this->matchLabels($realNode, $labels, $labelNodeName, $parentNodeName);
            }
        }
    }

    /**
     *
     *
     */
    public function hasExec($realNodes)
    {
        foreach ($realNodes as $realNode) {
            if ($realNode->setup) {
                return true;
            } elseif ($realNode->teardown) {
                return true;
            } else {
                return $this->hasExec($realNode);
            }
        }

        return false;
    }

    /**
     *
     */
    protected function isRealNodeInPlace($realNode, $validNodes)
    {
        $realNodeName = $realNode->getName();

        if (isset($validNodes[$realNodeName]) || $realNodeName == 'userstory') {
            return true;
        // TODO: Multiple <userstory>, <given>, <when>, ... tags
        // } elseif ($realNodeName == 'userstory') {
        //     return true;
        } else {
            return false;
        }
    }

    /**
     *
     *
     */
    protected function isRealNodeValid($realNode, $validNodes, $status)
    {
        if ($status) {
            return $status;
        } else {
            $validNodeChildren = array();
            $status = false;
            $realNodeName = $realNode->getName();

            foreach ((array) $validNodes as $validNodeName => $validNode) {
                // TODO: Does it ever get to here?
                if ($validNodeName == $realNodeName) {
                    return true;
                } elseif (isset($validNodes[$validNodeName]['children']) && is_array($validNodes[$validNodeName]['children'])) {
                    $validNodeChildren = $validNodes[$validNodeName]['children'];
                    $nextValidKeys = \array_keys($validNodeChildren);

                    if (\in_array($realNodeName, $nextValidKeys)) {
                        return true;
                    }
                }
            }

            return $this->isRealNodeValid($realNode, $validNodeChildren, $status);
        }
    }

    /**
     *
     *
     */
    protected function getValidNodeKeys($validNodes, $validNodeKeys = array())
    {
        foreach ((array) $validNodes as $validNodeName => $validNode) {
            $validNodeKeys[] = $validNodeName;

            if (isset($validNodes[$validNodeName]['children']) && is_array($validNodes[$validNodeName]['children'])) {
                $validNodeChildren = $validNodes[$validNodeName]['children'];
                $validNodeKeys = $this->getValidNodeKeys($validNodeChildren, $validNodeKeys);
            }
        }

        return $validNodeKeys;
    }

    /**
     *
     */
    protected function suggestNodes($realNode, $validNodes)
    {
        $validNodeKeys = $this->getValidNodeKeys($this->validXmlStructure);

        if (in_array($realNode->getName(), $validNodeKeys)) {
            $possibleValidNodes = \array_keys($validNodes);

            if ($possibleValidNodes) {
                $possibleValidNodeString = '<' . \implode('>, <', $possibleValidNodes) . '>';

                $this->addError($realNode, 'The <' . $realNode->getName() . '> node is in the wrong place. Only the nodes ' . $possibleValidNodeString . ' are valid here');

                return true;
            }
        }

        return false;
    }

    /**
     *
     */
    protected function getNextValidNodes($realNode, $validNodes)
    {
        $realNodeName = $realNode->getName();

        if (isset($validNodes[$realNodeName]['children'])) {
            return $validNodes[$realNodeName]['children'];
        // TODO: Multiple <userstory>, <given>, <when>, ... tags
        //} elseif ($realNodeName == 'userstory') {
        //    return $validNodes;
        } else {
            return array();
        }
    }

    /**
     * TODO: Loop through all valid nodes until found (instead of passing $validNodeChildren)
     */
    protected function isBottomNode($realNode, $validNodeChildren)
    {
        $validNodeChildren = $this->getNextValidNodes($realNode, $validNodeChildren);

        if ($validNodeChildren) {
            return false;
        } else {
            return true;
        }

        return false;

    }

    /**
     *
     */
    protected function addError($simpleXmlNode, $errorReason)
    {
        // TODO: Do this check everywhere else too where there is dom_import_simplexml()
        if ($simpleXmlNode instanceof \AutomateYourTests\Model\Config\ExtendedXMLElement) {
            $domNode = \dom_import_simplexml($simpleXmlNode);

            $this->validationErrors[$domNode->getLineNo()] = $domNode->getLineNo() . ': ' . \realpath($domNode->baseURI) . ': ' . $errorReason;
        } else {
            $uniqueKey = '???:' . rand(0, 100000);

            $this->validationErrors[$uniqueKey] = '???: ' . $errorReason;
        }
    }

    /**
     *
     */
    protected function reportValidationErrors()
    {
        if ($this->validationErrors) {
            \ksort($this->validationErrors, SORT_NUMERIC);

            $errorPhrase = "Your test suite XML is invalid:\n" . implode(".\n", $this->validationErrors) . ".";

            throw new \Exception($errorPhrase);
        }
    }
}
