<?php
namespace AutomateYourTests\Controller;

/**
 * The following example runs small proof-of-concept test scenario on a dedicated Wordpress demo web application that I
 * have set up especially for this purpose. Each XML title="" attribute below documents the exact feature of the
 * testing software that this demo tries to teach. Enough words ... make sure you have Internet access, then simply
 * change into the directory where you extracted the “AutomateYourTests” ZIP and execute one of the following
 * commands and see what happens ...
 *
 * @example     php run.php testsuite="Examples/1_3_hello_world.xml"
 * @example     php run.php testsuite="Examples/1_3_hello_world.xml" from="20" to="25"
 * @example     php run.php testsuite="Examples/1_3_hello_world.xml" verbose="failure" to="failure"
 *
 * @link        http://automateyourtests.com
 * @version     1.2.2
 * @author      Marc Daniel Dammann <howto@automateyourtests.com>
 * @copyright   2016 Marc Daniel Dammann
 * @license     GPL-3.0 <https://opensource.org/licenses/GPL-3.0>
 *
 * Copyright 2016 Marc Daniel Dammann
 * This file is part of AutomateYourTests. AutomateYourTests is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version. AutomateYourTests is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with AutomateYourTests. If not, see http://www.gnu.org/licenses/.
 */

class Base {
    /**
     * Class vars
     */
    protected $config = array();
    protected $testsuite = '';
    protected $baseurl = '';
    protected $model = null;
    protected $view = null;
    protected $controller = null;

    /**
     * Get set up
     */
    public function __construct($commandLineOptions)
    {
        // Be safe
        if (version_compare(PHP_VERSION, '5.4.0') < 0) {
            echo('Your PHP version is ' . PHP_VERSION . ', but AutomateYourTests requires at least 5.4.0');
        } else {
            // Avoid php.ini date.timezone warning
            date_default_timezone_set('Etc/UTC');

            // Load application classes
            $this->registerAutoloader();
            $this->instantiateAppClasses();

            // Better safe than sorry
            $this->model->config->environmentor->checkSystemRequirements();

            // Get command line arguments
            $commandLineOptions = $this->parseCommandLineOptions($commandLineOptions);
            $this->sanitizeCommandLineOptions($commandLineOptions);

            // Safeguard against file-system non-sense
            // TODO: Circumvent is_file() and is_readable(), if filter_var($realNode, FILTER_VALIDATE_URL) !== false (but also check the php.ini allow_url_fopen AND allow_url_include php.ini settings)
            if (! $this->testsuite) {
                throw new \Exception('The PHP script that you have executed is missing the testsuite="" argument (what automated tests do you want to run?)');
            } elseif (! @is_readable($this->testsuite)) {
                if (! file_exists($this->testsuite)) {
                    throw new \Exception('The specified test suite config path "' . $this->testsuite . '" does not exist');
                } else {
                    throw new \Exception('The specified test suite config "' . $this->testsuite . '" does not have READ permission for this user (or your php.ini setting "open_basedir = ' . \ini_get('open_basedir') . '" path excludes it)');
                }
            } elseif (! is_file($this->testsuite)) {
                throw new \Exception('The specified test suite config path "' . $this->testsuite . '" is not a file (maybe a directory?)');
            } else {
                // Load config XML
                // TODO: Should we apply realpath() here first?
                $config = $this->loadXmlFilePath();

                // Make sure XML structure is (preliminary) sound
                if (! $this->controller->debugger->isDebug()) {
                    $this->model->config->validator->checkSome($config);
                }

                // If <setup> or <teardown>
                if ($this->model->config->validator->hasExec($config)) {
                    $this->model->config->environmentor->checkExec();
                }

                // Allow command line argument to override 'baseurl'
                $config = $this->model->config->inheritor->injectBaseUrl($config, $this->baseurl);

                // TODO: WORKAROUND: Pro-forma root to recurse over
                $configRoot = new \stdClass();
                $configRoot->testsuite = array($config);

                $configRoot = $this->model->config->inheritor->populate($configRoot);

                // Prepend 'url' with 'baseurl'
                // TODO: Move this into the inheritor?
                $this->model->config->normaliser->completeUrls($configRoot);
                $this->model->config->normaliser->completeUserImpersonationUrls($config);

                // Clean up strange SimpleXml comments
                $this->model->config->normaliser->removeXmlComments($config);

                // Make sure once again XML structure is (still finally) sound
                if (! $this->controller->debugger->isDebug()) {
                    $this->model->config->validator->checkAll($config);
                }

                // Reload final restructured XML
                //$config = $this->reloadXml($config);

                if ($this->controller->debugger->isDebug()) {
                    print "<pre>Parsed config:\n\n"; print_r($configRoot); print "</pre>\n";
                    print "<pre>Parsed arguments:\n\n"; print_r($commandLineOptions); print "</pre>\n";
                    exit;
                }

                // Set config across classes
                $this->setConfig($config);
                $this->controller->urlCaller->setConfig($config);
                $this->controller->logins->setConfig($config);
                $this->model->fixture->setConfig($config);
                $this->view->logger->setConfig($config);
                $this->controller->cascader->setConfig($config);
            }
        }
    }

    /**
     * Instantiate other MVC models.
     * You may override this method in order to inject (some of) your own versions of application classes.
     */
    protected function instantiateAppClasses()
    {
        $this->model                        = new \stdClass();
        $this->model->config                = new \stdClass();
        $this->model->config->environmentor = new \AutomateYourTests\Model\Config\Environmentor();
        $this->model->config->normaliser    = new \AutomateYourTests\Model\Config\Normaliser();
        $this->model->config->inheritor     = new \AutomateYourTests\Model\Config\Inheritor($this->model->config->normaliser);
        $this->model->config->validator     = new \AutomateYourTests\Model\Config\Validator();
        $this->model->asserter              = new \AutomateYourTests\Model\Asserter();
        $this->model->counter               = new \AutomateYourTests\Model\Counter($this->model->config->normaliser);
        $this->model->fixture               = new \AutomateYourTests\Model\Fixture();

        $this->view                         = new \stdClass();
        $this->view->logger                 = new \AutomateYourTests\View\Logger($this->model->counter, $this->model->fixture);

        $this->controller                   = new \stdClass();
        $this->controller->debugger         = new \AutomateYourTests\Controller\Debugger();
        $this->controller->urlCaller        = new \AutomateYourTests\Controller\UrlCaller();
        $this->controller->logins           = new \AutomateYourTests\Controller\LoginUser($this->controller->urlCaller);
        $this->controller->cascader         = new \AutomateYourTests\Controller\Cascader($this->testsuite, $this->model, $this->view, $this->controller);
    }

    /**
     * Get command line arguments.
     * WARNING: Anything that gets passed as "name instead of "name=value" has a vakue of integer 1 after this.
     */
    protected function parseCommandLineOptions($argv)
    {
        $parsed = array();

        array_shift($argv);

        foreach ($argv as $a) {
            $eq = strpos($a, '=');

            if ($eq !== false) {
                $parsed[substr($a, 0, $eq)] = substr($a, $eq + 1);
            } elseif (! isset($parsed[$a])){
                $parsed[$a] = true;
            }
        }

        return $parsed;
    }

    /**
     * Get command line arguments
     */
    /* protected function parseCommandLineOptions()
    {
        $shortOptions = '';
        //$shortOptions  = 'testsuite:';  // Required value
        //$shortOptions .= 'baseurl::';   // Optional value
        //$shortOptions .= 'debug::';     // Optional value

        $longOptions = array(
            'testsuite:',   // Required value
            'baseurl::',    // Optional value
            'debug::',      // Optional value
        );
        //$longOptions  = array();

        return getopt($shortOptions, $longOptions);
    } */

    /**
     * Sanitize and assign command line arguments.
     * WARNING: Anything that gets passed as "name instead of "name=value" has a vakue of integer 1.
     * TODO: Move to Validator()
     */
    protected function sanitizeCommandLineOptions($options)
    {
        // Defaults
        $testsuite = '';
        $verbose = false;
        $interactive = false;
        $from = 0;
        $to = 0;
        $debug = false;
        $baseurl = '';


        // TODO: Get \realpath($options['testsuite'] in here too?
        if (isset($options['testsuite'])) {
            if (! empty($options['testsuite']) && is_string($options['testsuite'])) {
                $testsuite = \filter_var($options['testsuite'], FILTER_SANITIZE_STRING);
            } else {
                throw new \Exception('The command line argument "testsuite" must be a string');
            }
        } else {
            throw new \Exception('The command line argument "testsuite" must be the path of the test suite XML that contains your automated tests');
        }

        if (isset($options['verbose'])) {
            if ($options['verbose'] === true || $options['verbose'] === 1) {
                $verbose = 'all';
            } elseif ($options['verbose'] === 'all' || $options['verbose'] === 'failure') {
                $verbose = $options['verbose'];
            } else {
                throw new \Exception('The command line argument "verbose" must either be verbose="all" or verbose="pass" or verbose="failure"');
            }
        }

        if (isset($options['interactive'])) {
            if ($options['interactive'] === true || $options['interactive'] === 1) {
                $interactive = 'all';
            } elseif ($options['interactive'] === 'all' || $options['interactive'] === 'failure') {
                $interactive = $options['interactive'];
            } else {
                throw new \Exception('The command line argument "interactive" must either be interactive="all" or interactive="pass" or interactive="failure"');
            }
        }

        if (isset($options['from'])) {
            if (\filter_var($options['from'], FILTER_VALIDATE_INT)) {
                if ($options['from'] >= 0) {
                    $from = \filter_var($options['from'], FILTER_SANITIZE_NUMBER_INT);
                } else {
                    throw new \Exception('The command line argument "from" must be a positive integer');
                }
            } else {
                throw new \Exception('The command line argument "from" must be an integer');
            }
        }

        if (isset($options['to'])) {
            if (\filter_var($options['to'], FILTER_VALIDATE_INT)) {
                if ($options['to'] >= 0) {
                    $to = \filter_var($options['to'], FILTER_SANITIZE_NUMBER_INT);
                } else {
                    throw new \Exception('The command line argument "to" must be a positive integer');
                }
            } elseif ($options['to'] === 'failure') {
                $to = $options['to'];
            } else {
                throw new \Exception('The command line argument "to" must be an integer');
            }
        }

        if (isset($options['debug'])) {
            if ($options['debug'] === true || $options['debug'] === 1) {
                $debug = true;
            } else {
                $debug = false;
            }
        }

        if (isset($options['baseurl'])) {
            if (! empty($options['baseurl']) && is_string($options['baseurl'])) {
                $baseurl = \filter_var($options['baseurl'], FILTER_SANITIZE_STRING);
            } else {
                throw new \Exception('The command line argument "baseurl" must be a string');
            }
        }

        // Set class vars
        $this->testsuite = $testsuite;
        $this->baseurl = $baseurl;
        $this->controller->debugger->setVerbose($verbose);
        $this->controller->debugger->setInteractive($interactive);
        $this->controller->debugger->setFrom($from);
        $this->controller->debugger->setTo($to);
        $this->controller->debugger->setDebug($debug);

        return $options;
    }

    /**
     * Load config XML from disk into array
     */
    protected function loadXmlFilePath()
    {
        \libxml_use_internal_errors(true);

        // Allow <![CDATA[]]> content
        $libXmlOptions = LIBXML_NOCDATA;
        //$libXmlOptions = LIBXML_NOERROR;
        //$libXmlOptions = LIBXML_XINCLUDE;

        try {
            // TODO: Allow $data_is_url to be true for URLs!? But then beware of allow_url_fopen php.ini setting!
            $config = new \AutomateYourTests\Model\Config\ExtendedXMLElement($this->testsuite, $libXmlOptions, true);
            //$config->encoding='UTF-8';
            $config->enhanceXmlParsing();
        } catch (\Exception $e) {
            $errorStrings = array();

            $errors = libxml_get_errors();

            foreach ($errors as $error) {
                // Ignore LIBXML_ERR_WARNING
                if ($error->level == LIBXML_ERR_ERROR || $error->level == LIBXML_ERR_FATAL) {
                    $errorStrings[] = $error->line . ': ' . \realpath($error->file) . ' has the problem: ' . str_replace("\n", "", $error->message);
                }
            }

            \libxml_clear_errors();

            if ($errorStrings) {
                throw new \Exception("XML could not be parsed:\n" . implode("\n", $errorStrings));
            }
        }

        return $config;
    }
    
    /**
     * Re-load config XML into SimpleXML after we have modified it (inheritance, etc).
     * This is so that error messages can maintain track of the the correct line number
     * within the modified XML "file".
     */
    /* protected function reloadXml($config)
    {
        // Allow <![CDATA[]]> content
        $libXmlOptions = LIBXML_NOCDATA;
        $config = new \AutomateYourTests\Model\Config\ExtendedXMLElement($config->asXML(), $libXmlOptions, false);
        //$config->encoding='UTF-8';

        return $config;
    } */
    /**
     * Autoload application classes from the AutomateYourTests namespace (on demand when being instantiated).
     */
    protected function registerAutoloader() {
        \spl_autoload_register(function ($className) {
            $absoluteAppClassPath = __DIR__ . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR . str_replace('\\', DIRECTORY_SEPARATOR, $className) . '.php';

            // Safeguard against any file system non-sense
            //if (! @\is_executable($absoluteAppClassPath)) { // @ protects against PHP warnings
            //    throw new \Exception('The application class file "' . $absoluteAppClassPath . '" does not have the required EXECUTE permission (or your php.ini setting "open_basedir = ' . \ini_get('open_basedir') . '" path excludes it)');
            if (! \is_file($absoluteAppClassPath)) {
                throw new \Exception('The application class file "' . $absoluteAppClassPath . '" does not exist on disk (or your php.ini setting "open_basedir = ' . \ini_get('open_basedir') . '" path excludes it)');
            } else {
                require_once($absoluteAppClassPath);
            }
        });
    }

    /**
     * Delegate to Cascader() to loop through all tests
     */
    public function run()
    {
        $this->controller->cascader->runTests();
    }
    
    /**
     * Return verbose log so that the user retains the flexibility to do whatever they want with it
     */
    public function getLog($outputTemplate = 'standard')
    {
        if ($outputTemplate == 'standard') {
            if ($this->controller->debugger->isInteractive()) {
                $log = $this->view->logger->getInteractiveLog();
            } else {
                $log = $this->view->logger->getLog();
            }
        } elseif ($outputTemplate == 'summary') {
            if ($this->controller->debugger->isInteractive()) {
                $log = $this->view->logger->getInteractiveSummary($this->testsuite);
            } else {
                $log = $this->view->logger->getSummary($this->testsuite);
            }
        } else {
            throw new \Exception('The output template "' . $outputTemplate . '" passed to getLog() is not valid. It must be either "standard" or "summary" or left empty');
        }

        return $log;
    }

    /**
     *
     */
    protected function setLog()
    {
        $this->view->logger->set('');
    }

    /**
     * Return exit code. It is up them then, what they want to
     * do with it (version control, continuous integration, ...)
     */
    public function getExitCode()
    {
        // TODO: Give access to failureCount itself through templating? (Or better: $this->getFailureCount() ???)
        if ($this->view->logger->failureCount == 0) {
            return 0;
        } else {
            return 1;
        }
    }

    /**
     * Set config class var
     */
    public function setConfig($config = array())
    {
        $this->config = $config;
    }
}

// TODO: Move this elsewhere
class TerminateIfFailException extends \Exception {}
